;******************************************************************************
; file:      main.asm
; content:   evaluation project for sf20b processors
;            prints "hello world"
;            runs sieve of Eratosthenes and print prime numbers up to 999
; author:    Martin Raubuch
; created:   22Dec2014
; modified:  22Dec2014
; Property of RACORS GmbH, Ginsterweg 1a, 85598 Baldham, Germany
;******************************************************************************
; ABI Conventions, allocation order
; R8           used as stack/frame pointer
; RF-RC        pointer argument passing, may be destroyed by called function
; R0-R3        data argument passing, may be destroyed by called function
; R0           return value
; R4-R7/R9-RB  if modified must be saved/restored by called function	
;******************************************************************************

;******************************** Reset Entry *********************************
; typical reset entry code:
; - sets IA to the program start address
; - sets the interrupt vector table start address (not used in this project)
; - initializes the stack pointer R8
; - executes an rtir to leave the interrupt state abd jump to 'PrgStart'
;******************************************************************************
   .text                        ; switch to text segment
   .org   0x0000                ; reset start address
   mtsr   PrgStart,IA 		; IA = &PrgStart
   move   IrVecTab&0xff,R0      ; R0 = &IrVecTab
   addh   IrVecTab>>8,R0	; R0 = &IrVecTab
   mtsr   R0,CS         	; set interrupt vector table address
   move   TopOfStk&0xff,R8	; R8 = TopOfStk
   addh   TopOfStk>>8,R8	; R8 = TopOfStk
   rtir				; return from interrupt
;******************************* Program Start ********************************
PrgStart:
   jpsr   PrintHelloWorld	; PrintHelloWorld()
   jpsr   Sieve         	; Sieve(), prime number search up to 1000
   jpsr   PrintPrimes         	; PrintPrimes(), print prime numbers
;******************************** Program End *********************************
   .data			; data section
EndStr:
   .asciz  "Program End"	; 0-terminated string
   .text			; text section
   jpsr   simcycle		; simcycle(), print simulation cycle number
   move   EndStr&0xff,RF	; RF = &EndStr
   addh   EndStr>>8,RF  	; RF = &EndStr
   jpsr   puts			; puts(RF)
   move   '\n',R0		; R0 = '\n', newline
   jpsr   putchar		; putchar(R0)
ProgEnd:
   bral   ProgEnd		; endless loop at program end
;***************************** Print Hello World ******************************
; writes string "Hello World" + CR (carriage return) to stdout
; note that S (register SA) needs to be saved on the stack because a nested
; subroutine is called
;******************************************************************************
   .data			; data section
HloWldStr:
   .asciz  "Hello World"	; 0-terminated string
   .text			; text section
PrintHelloWorld:
   stsh   "S",-(R8)		; save registers to stack
   jpsr   simcycle		; simcycle(), print simulation cycle number
   move   HloWldStr&0xff,RF	; RF = &HloWldStr
   addh   HloWldStr>>8,RF	; RF = &HloWldStr
   jpsr   puts			; puts(RF)
   move   '\n',R0		; R0 = '\n', newline
   jpsr   putchar		; putchar(R0)
   ldsh   (R8)+,"S"		; restore registers from stack
   rtsr				; quit
;************************* Sieve (prime number search) ************************
; searches prime numbers up to 999 using the sieve of Eratosthenes algorithm
;******************************************************************************
   .data			; data section
SieveTxt:
   .asciz  "Searchig Prime Numbers ..."	; 0-terminated string
   .text			; text section
Sieve: 
   stsh   "S",-(R8)		; save registers to stack
   jpsr   simcycle		; simcycle(), print simulation cycle number
   move   SieveTxt&0xff,RF	; RF = &SieveTxt
   addh   SieveTxt>>8,RF	; RF = &SieveTxt
   jpsr   puts			; puts(RF)
   move   '\n',R0		; R0 = '\n', newline
   jpsr   putchar		; putchar(R0)
   move   SieveBuf&0xff,RF      ; buffer start address
   addh   SieveBuf>>8,RF        ; buffer start address
   move   0x01,R0               ; init value
   addh   0x01,R0               ; init value
   mtsr   125,LC                ; loop count, init 1000 bytes
1: stsh   R0,(RF)+		; clear 4 bytes
   stsh   R0,(RF)+		; clear 4 bytes
   stsh   R0,(RF)+		; clear 4 bytes
   stsh   R0,(RF)+		; clear 4 bytes
   brlc   1b			; LC--, if != 0 loop back to 1b
   move   SieveBuf&0xff,RF      ; buffer start address
   addh   SieveBuf>>8,RF        ; buffer start address
   move   0,R0		        ; cleared entry value
   move   2,R1                  ; outer loop counter
   move   1000&0xff,RD		; RD = 1000, max product
   addh   1000>>8,RD		; RD = 1000, max product
1: move   2,RC                  ; inner loop counter
2: mult   R1,RC,R3		; R3 = R1*RC
   addt   1,RC			; RC += 1
   mult   R1,RC,R2		; R2 = R1*RC
   addt   1,RC			; RC += 1
   stbt   R0,(R3,RF)	        ; clear non prime entry
   comp   RD,R2			; product > 1000
   stbt   R0,(R2,RF)	        ; clear non prime entry
   brcr   2b			; if not loop back to 2b
   addt   1,R1			; R1 += 1
   comp   32,R1			; R1 == 32, ~SQRT(1000)
   brcr   1b			; if not loop back to 1b
   ldsh   (R8)+,"S"		; restore registers from stack
   rtsr				; quit
;***************************** Print Prime Numbers ****************************
; prints the numbers in 'SieveBuf' as decimal numbers
;******************************************************************************
   .data			; data section
PrintTxt:
   .asciz  "Numbers found:"	; 0-terminated string
   .text			; text section
PrintPrimes: 
   stsh   "S45B",-(R8)		; save registers to stack
   jpsr   simcycle		; simcycle(), print simulation cycle number
   move   PrintTxt&0xff,RF	; RF = &PrintTxt
   addh   PrintTxt>>8,RF	; RF = &PrintTxt
   jpsr   puts			; puts(RF)
   move   '\n',R0		; R0 = '\n', newline
   jpsr   putchar		; putchar(R0)
   move   0,R0			; R0 = 0
   stbt   R0,NumStrBuf+3	; NumStrBuf[3] = 0, to avoid compare to undef.
   move   SieveBuf&0xff,RB      ; buffer start address
   addh   SieveBuf>>8,RB        ; buffer start address
   move   2,R4			; R4 = 2, start with 2
   move   1000&0xff,R5		; R5 = 1000, max product
   addh   1000>>8,R5		; R5 = 1000, max product
1: ldbt   (R4,RB),R0		; R0 = next entry from SieveBuf
   move   3,R1			; R1 = 3, #of digits
   comp   1,R0			; R0==1 ?, prime number
   move   R4,R0			; R0 = index -> prime number
   brzr   3f			; if yes continue at 3f
2: addt   1,R4			; R4++
   comp   R5,R4  		; R4==R5 ? end of buffer
   brnz   1b			; if not loop back to 1b
   ldsh   (R8)+,"S45B"		; restore registers from stack
   rtsr				; quit
3: move   NumStrBuf,RF		; RF = &NumStrBuf, number string buffer
   jpsr   DecNumToTxt		; DecNumToTxt(R0,R1,RF)
   jpsr   simcycle		; simcycle(), print simulation cycle number
   move   NumStrBuf,RF		; RF = &NumStrBuf, number string buffer
   jpsr   puts			; puts(RF)
   move   '\n',R0		; R0 = '\n', newline
   jpsr   putchar		; putchar(R0)
   bral   2b			; loop back to 2b
;********************* Decimal Number to Text conversion **********************
; converts a 16-bit integer to an ascii string, max 5 character
; leading zeroes are filled with blank ' ' characters
; input: R0 number
; input: R1 #of digits (max 5)
; input: RF pointer to text string buffer
;******************************************************************************
DecNumToTxt:
   stsh   "45ABS",-(R8)    	; save registers
   move   R1,R4			; save #of digits
   move   R1,R5 		; save #of digits
   move   RF,RA			; save string pointer
   move   R8,RB			; RB = digit string buffer (on stack)
1: move   10,R1			; R1 = divisor = 10
   jpsr   IntDiv1616		; IntDiv1616(R0,R1)
   addt   0x30,R0		; R0 = remainder digit ASCII code
   subf   1,RB			; RB--
   stbt   R0,(RB)		; store in string buffer
   move   R1,R0			; R0 = next divident = this quotient
   subf   1,R4			; R4--
   brnz   1b			; if != 0 loop back to 1b
   move   0x20,R1		; R1 = 0x20 = ' '
1: ldbt   (RB)+,R0		; R0 = ASCII character
   comp   0x30,R0		; character == '0' (leading)
   brnz   2f			; if not continue at 2f
   stbt   R1,(RA)+  		; store a ' ' character
   subf   1,R5  		; all digits written ?
   comp   1,R5			; R5 == 1 ?
   brnz   1b			; if not loop back
1: ldbt   (RB)+,R0		; R0 = ASCII character
2: stbt   R0,(RA)+		; write to string buffer
   subf   1,R5  		; all digits written ?
   brnz   1b			; if not loop back
   ldsh   (R8)+,"45ABS"   	; restore registers
   rtsr				; quit
